<?php

namespace Clue\PharComposer;

use UnexpectedValueException;
use InvalidArgumentException;
use Herrera\Box\Box;
use Herrera\Box\StubGenerator;
use RecursiveIteratorIterator;
use RecursiveDirectoryIterator;
use FilesystemIterator;

class PharComposer
{
    private $pathProject;
    private $package;
    private $main = null;
    private $target = null;

    public function __construct($path)
    {
        $path = realpath($path);

        $this->package = json_decode(file_get_contents($path), true);
        if ($this->package === null) {
            var_dump(json_last_error(), JSON_ERROR_SYNTAX);
            throw new InvalidArgumentException('Unable to parse given path "' . $path . '"');
        }

        $this->pathProject = dirname($path) . '/';
    }

    public function getTarget()
    {
        if ($this->target === null) {
            if (isset($this->package['name'])) {
                // skip vendor name from package name
                $this->target = substr($this->package['name'], strpos($this->package['name'], '/') + 1);
            } else {
                $this->target = basename($this->pathProject);
            }
            $this->target .= '.phar';
        }
        return $this->target;
    }

    public function setTarget($target)
    {
        $this->target = $target;
        return $this;
    }

    public function getMain()
    {
        if ($this->main === null) {
            if (isset($this->package['bin'])) {
                foreach ($this->package['bin'] as $path) {
                    $path = $this->getAbsolutePathForComposerPath($path);
                    if (!file_exists($path)) {
                        throw new UnexpectedValueException('Bin file "' . $path . '" does not exist');
                    }
                    $this->main = $path;
                    break;
                }
            }
        }
        return $this->main;
    }

    public function setMain($main)
    {
        $this->main = $main;
        return $this;
    }

    /**
     * base project path. all files MUST BE relative to this location
     *
     * @return string
     */
    public function getBase()
    {
        return $this->pathProject;
    }

    public function build()
    {
        // var_dump($this->package);

        $target = $this->getTarget();
        echo 'Start creating "'.$target.'"...' . PHP_EOL;
        if (file_exists($target)) {
            echo 'Remove existing file...';
            if(unlink($target) === false) {
                throw new UnexpectedValueException('Unable to remove existing phar archive "'.$target.'"');
            } else {
                echo ' ok'. PHP_EOL;
            }
        }
        $box = Box::create($target);

        $main = $this->getMain();
        if ($main === null) {
            echo 'WARNING: No main bin file defined! Resulting phar will NOT be executable' . PHP_EOL;
        } else {
            $this->addFile($box, $main);
            // $box->addFile($main);
            $box->getPhar()->setStub(
                StubGenerator::create()
                ->alias('default.phar')
                ->index($this->getPathLocalToBase($main))
                ->generate()
            );
        }

        if (isset($this->package['autoload'])) {
            if (isset($this->package['autoload']['psr-0'])) {
                foreach ($this->package['autoload']['psr-0'] as $namespace => $path) {
                    if ($namespace !== '') {
                        // namespace given, add it to the path
                        $namespace = str_replace('\\', '/', $namespace);

                        if ($path === '') {
                            // namespace in project root => namespace is path
                            $path = $namespace;
                        } else {
                            // namespace in sub-directory => add namespace to path
                            $path = rtrim($path, '/') . '/' . $namespace;
                        }
                    }

                    // TODO: this is not correct actually... should work for most repos nevertheless
                    // TODO: we have to take target-dir into account

                    $this->addDirectory($box, $this->getAbsolutePathForComposerPath($path));
                }
            }

            if (isset($this->package['autoload']['classmap'])) {
                foreach($this->package['autoload']['classmap'] as $path) {
                    $path = $this->getAbsolutePathForComposerPath($path);
                    if (is_dir($path)) {
                        $this->addDirectory($box, $path);
                    } else {
                        $this->addFile($box, $path);
                    }
                }
            }

            if (isset($this->package['autoload']['files'])) {
                foreach($this->package['autoload']['classmap'] as $path) {
                    $this->addFile($box, $this->getAbsolutePathForComposerPath($path));
                }
            }
        }

        $vendor = 'vendor';
        if (isset($this->package['config']['vendor-dir'])) {
            $vendor = $this->package['config']['vendor-dir'];
        }
        $this->addDirectory($box, $this->getAbsolutePathForComposerPath($vendor));
    }

    private function addDirectory(Box $box, $dir)
    {
        $dir = rtrim($dir, '/') . '/';

        echo 'adding "' . $dir .'" as "' . $this->getPathLocalToBase($dir).'"...';
        $iterator = new RecursiveIteratorIterator(
            new RecursiveDirectoryIterator(
                $dir,
                FilesystemIterator::KEY_AS_PATHNAME
                | FilesystemIterator::CURRENT_AS_FILEINFO
                | FilesystemIterator::SKIP_DOTS
            )
        );

        $box->buildFromIterator($iterator, $this->pathProject);
        echo ' ok' . PHP_EOL;
    }

    private function addFile(Box $box, $file)
    {
        $local = $this->getPathLocalToBase($file);
        echo 'adding "' . $file .'" as "' . $local.'"...';
        $box->addFile($file, $local);
        echo ' ok' . PHP_EOL;
    }

    /**
     *
     * @param string $path
     * @return string
     */
    private function getAbsolutePathForComposerPath($path)
    {
        // return $path;
        return $this->pathProject . rtrim($path, '/');
    }

    private function getPathLocalToBase($path)
    {
        if (strpos($path, $this->pathProject) !== 0) {
            throw new UnexpectedValueException('Path "' . $path . '" is not within base project path "' . $this->pathProject . '"');
        }
        return substr($path, strlen($this->pathProject));
    }
}
